//! This module contains the whole engine that manages everything.

pub mod bbox;
pub mod character;
pub mod controls;
pub mod event;
pub mod map;
pub mod math;
pub mod physics;
pub mod scene;
pub mod texture;
pub mod vector;

use std::cell::RefCell;
use std::rc::Rc;

use wasm_bindgen::prelude::*;
use wasm_bindgen::JsCast;

use crate::engine::event::{Key, Keyboard};
use crate::engine::texture::TextureManager;
use crate::{error_js, Result};

macro_rules! unwrap {
    ($t: expr) => {{
        match $t {
            None => return Err(JsValue::from_str("cannot unwrap on none")),
            Some(x) => x,
        }
    }};
}

/// Our game engine.
pub struct Engine {
    /// The inner engine.
    ///
    /// We need Rc<RefCell> in order to deal with events.
    pub inner: Rc<RefCell<InnerEngine>>,

    /// The web page document.
    pub document: Rc<web_sys::Document>,

    /// The canvas rendering context.
    ///
    /// We keep a reference so that we can easily render things.
    pub context: Rc<web_sys::CanvasRenderingContext2d>,
}

impl Engine {
    /// Creates a new engine.
    pub fn new() -> Result<Engine> {
        let window = unwrap!(web_sys::window());
        let document = unwrap!(window.document());
        let canvas = unwrap!(document.get_element_by_id("canvas"));
        let canvas: web_sys::HtmlCanvasElement = canvas.dyn_into::<web_sys::HtmlCanvasElement>()?;

        canvas.set_width(1920);
        canvas.set_height(1080);

        let context = canvas
            .get_context("2d")?
            .unwrap()
            .dyn_into::<web_sys::CanvasRenderingContext2d>()?;

        let inner = InnerEngine::new(&document)?;

        let document = Rc::new(document);
        let context = Rc::new(context);

        Ok(Engine {
            inner: Rc::new(RefCell::new(inner)),
            document,
            context,
        })
    }

    /// Clones the engine.
    ///
    /// It is made to be efficient, cloning only Rcs.
    pub fn clone(&self) -> Engine {
        Engine {
            inner: self.inner.clone(),
            document: self.document.clone(),
            context: self.context.clone(),
        }
    }

    /// Starts the engine.
    pub fn start(&self) -> Result<()> {
        let clone = self.clone();
        let cb = Closure::<dyn FnMut(_)>::new(move |_event: web_sys::Event| {
            if let Err(e) = clone.run() {
                error_js!(e);
            }
        });

        web_sys::window()
            .unwrap()
            .request_animation_frame(cb.as_ref().unchecked_ref())
            .unwrap();

        cb.forget();

        Ok(())
    }

    /// Launches a loop of the engine, and schedules the next one.
    pub fn run(&self) -> Result<()> {
        // Perform update
        self.update()?;

        // Perform render
        self.render()?;

        // Schedule next render
        let clone = self.clone();
        let cb = Closure::<dyn FnMut(_)>::new(move |_event: web_sys::Event| {
            if let Err(e) = clone.run() {
                error_js!(e);
            }
        });

        web_sys::window()
            .unwrap()
            .request_animation_frame(cb.as_ref().unchecked_ref())
            .unwrap();

        cb.forget();

        Ok(())
    }

    /// Triggers the update of the model of the engine.
    pub fn update(&self) -> Result<()> {
        let mut inner = self.inner.borrow_mut();

        // Manage events
        while let Some(_event) = inner.keyboard.pop() {
            // Nothing to do here yet
        }

        // Perform update
        let mut dx = 0.0;
        let mut dy = 0.0;

        if inner.keyboard.is_key_pressed(Key::ArrowDown) {
            dy += 5.0;
        }

        if inner.keyboard.is_key_pressed(Key::ArrowUp) {
            dy -= 5.0;
        }

        if inner.keyboard.is_key_pressed(Key::ArrowLeft) {
            dx -= 5.0;
        }

        if inner.keyboard.is_key_pressed(Key::ArrowRight) {
            dx += 5.0;
        }

        inner.x += dx;
        inner.y += dy;

        Ok(())
    }

    /// Performs the rendering of the engine.
    pub fn render(&self) -> Result<()> {
        let inner = self.inner.borrow();

        // Perform render
        self.context.clear_rect(0.0, 0.0, 1920.0, 1080.0);
        inner
            .textures
            .rusty
            .render(inner.x, inner.y, &self.context)?;

        Ok(())
    }
}

/// The data contained in our engine.
pub struct InnerEngine {
    /// The x position of the drawing.
    pub x: f64,

    /// The y position of the drawing.
    pub y: f64,

    /// The keyboard.
    pub keyboard: Keyboard,

    /// The texture manager.
    pub textures: TextureManager,
}

impl InnerEngine {
    /// Initializes the engine.
    pub fn new(document: &web_sys::Document) -> Result<InnerEngine> {
        Ok(InnerEngine {
            x: 0.0,
            y: 0.0,
            keyboard: Keyboard::new(document)?,
            textures: TextureManager::new()?,
        })
    }
}
